from jolly.config import config
from jolly.definitions import jolly
from .dynamo_table import DynamoTable
from jolly.models import JollyResource


class JollyResourcesTable(DynamoTable):

  RESOURCE_ID_COL = 'ResourceId'
  RESOURCE_TYPE_COL = 'ResourceType'
  INSTANCE_COL = 'Instance'

  def __init__(self):
    super(JollyResourcesTable, self).__init__(
      jolly.capitalize() + 'Resources' + config.ENV.capitalize()
    )

  def create_item(self, jolly_resource):
    return super(JollyResourcesTable, self).create_item(**{
      self.RESOURCE_ID_COL: jolly_resource.resource_id,
      self.RESOURCE_TYPE_COL: jolly_resource.resource_type,
      self.INSTANCE_COL: jolly_resource.instance,
    })



from jolly.services.jolly_resource_services.create_jolly_resource_service import CreateJollyResourceService
from jolly.models import JollyResource


def create_message(resource_id):
  return create_jolly_resource(JollyResource.MESSAGE_TYPE, resource_id)


def create_jolly_resource(resource_type, resource_id):
  # Initialize resource creation service.
  svc = CreateJollyResourceService(resource_type, resource_id)

  # Perform service.
  svc.perform()

  # Return the IP of the Jolly instance assigned to this resource.
  return svc.selected_instance_ip



from jolly.services.abstract_jolly_service import AbstractJollyService
from jolly.services.jolly_instance_services.select_jolly_instance_service import SelectJollyInstanceService
from jolly.services.jolly_instance_services.update_jolly_instance_load_service import UpdateJollyInstanceLoadService
from jolly.exp import *
from jolly.logger import logger
from jolly.models import JollyResource
from jolly.db import create_jolly_resource


class CreateJollyResourceService(AbstractJollyService):

  @property
  def selected_instance_ip(self):
    return self.jolly_instance.ip if self.jolly_instance else None

  def __init__(self, resource_type, resource_id, **kwargs):
    super(CreateJollyResourceService, self).__init__(exp=CreateJollyResourceException, **kwargs)
    self.resource_type = resource_type
    self.resource_id = resource_id
    self.jolly_instance = None
    self.jolly_resource = None

  def perform(self):
    logger.info('Creating new JollyResource(resource_id={}, resource_type={})...'.format(
      self.resource_id, self.resource_type))

    # Validate resource type.
    self._validate_resource_type()

    # Select a jolly instance for this resource.
    self._select_jolly_instance()

    # Create the jolly resource in the DB.
    self._create_jolly_resource()

    # Update the selected jolly instance's load in the DB to account for the new resource.
    self._update_jolly_instance_load()

    return self

  def _validate_resource_type(self):
    if self.resource_type not in JollyResource.resource_types:
      self.abort('Invalid resource type: {}'.format(self.resource_type))

  def _select_jolly_instance(self):
    try:
      self.jolly_instance = SelectJollyInstanceService(
        self.resource_type
      ).perform().jolly_instance
    except SelectJollyInstanceException as e:
      self.abort(e)
    except BaseException as e:
      self.abort('Unexpected error while selecting '
                 'JollyInstance for {}: {}'.format(self.resource_type, e))

  def _create_jolly_resource(self):
    # Create JollyResource model.
    self.jolly_resource = JollyResource(
      self.resource_id,
      self.resource_type,
      self.jolly_instance.instance,
    )

    # Write JollyResource to the DB.
    try:
      create_jolly_resource(self.jolly_resource)
    except JollyDBException as e:
      self.abort(e)
    except BaseException as e:
      self.abort('Unexpected error while creating JollyResource(resource_id={}, '
                 'resource_type={}): {}'.format(self.resource_id, self.resource_type, e))

  def _update_jolly_instance_load(self):
    try:
      UpdateJollyInstanceLoadService(
        self.jolly_instance,
        self.jolly_instance.load + self.jolly_resource.load,
      ).perform()
    except UpdateJollyInstanceException as e:
      self.abort(e)
    except BaseException as e:
      self.abort('Unexpected error while updating load of '
                 'JollyInstance {}: {}'.format(self.jolly_instance.instance, e))



from jolly.config import config
from jolly.definitions import jolly
from .dynamo_table import DynamoTable
from jolly.models import JollyInstance


class JollyInstancesTable(DynamoTable):

  INSTANCE_COL = 'Instance'
  IP_COL = 'IP'
  LOAD_COL = 'Load'

  def __init__(self):
    super(JollyInstancesTable, self).__init__(
      jolly.capitalize() + 'Instances' + config.ENV.capitalize()
    )

  def get_all(self):
    records = super(JollyInstancesTable, self).get_all()

    return [
      JollyInstance(
        record.get(self.INSTANCE_COL),
        record.get(self.IP_COL),
        record.get(self.LOAD_COL),
      )
      for record in records
    ]

  def update_load(self, instance, load):
    return super(JollyInstancesTable, self).update_item(
      Key={self.INSTANCE_COL: instance},
      AttributeUpdates={
        self.LOAD_COL: dict(Value=load, Action='PUT'),
      },
    )



from jolly.services.abstract_jolly_service import AbstractJollyService
from jolly.exp import *
from jolly.logger import logger
from jolly.db import get_jolly_instances


class SelectJollyInstanceService(AbstractJollyService):

  def __init__(self, resource_type, **kwargs):
    super(SelectJollyInstanceService, self).__init__(exp=SelectJollyInstanceException, **kwargs)
    self.resource_type = resource_type
    self.jolly_instances = []
    self.jolly_instance = None

  def perform(self):
    logger.info('Selecting JollyInstance for {}...'.format(self.resource_type))

    # Get all current jolly instances.
    self._get_jolly_instances()

    # Select the instance with the least load.
    self.jolly_instance = self._get_instance_with_least_load()

    return self

  def _get_jolly_instances(self):
    try:
      self.jolly_instances = get_jolly_instances()
    except JollyDBException as e:
      self.abort(e)
    except BaseException as e:
      self.abort('Unexpected error while querying '
                 'for all jolly instances: {}'.format(e))

  def _get_instance_with_least_load(self):
    return sorted(self.jolly_instances, key=lambda ji: ji.load)[0]


from jolly.services.abstract_jolly_service import AbstractJollyService
from jolly.exp import *
from jolly.logger import logger
from jolly.db import update_jolly_instance_load
from jolly.config import config


class UpdateJollyInstanceLoadService(AbstractJollyService):

  def __init__(self, jolly_instance, new_load, **kwargs):
    super(UpdateJollyInstanceLoadService, self).__init__(exp=UpdateJollyInstanceException, **kwargs)
    self.jolly_instance = jolly_instance
    self.new_load = new_load

  def perform(self):
    logger.info('Updating load of JollyInstance {} to {}...'.format(
      self.jolly_instance.instance, self.new_load))

    # Update load in the DB.
    self._update_jolly_instance()

    # Check if the Jolly cluster needs to scale.
    self._determine_if_cluster_should_scale()

    return self

  def _update_jolly_instance(self):
    try:
      update_jolly_instance_load(self.jolly_instance.instance, self.new_load)
    except JollyDBException as e:
      self.abort(e)
    except BaseException as e:
      self.abort('Unexpected error while updating load of JollyInstance '
                 '{}: {}'.format(self.jolly_instance.instance, e))

  def _determine_if_cluster_should_scale(self):
    if self._cluster_should_scale():
      self._schedule_scale_cluster_job()

  def _cluster_should_scale(self):
    return self.new_load > config.BALANCED_MAX_INSTANCE_LOAD

  def _schedule_scale_cluster_job(self):
    logger.info(
      'Balanced max instance load crossed on Jolly instance {} ({}/{}) -- '
      'Scheduling job to scale up the Jolly cluster...'.format(
        self.jolly_instance.instance, self.new_load, config.BALANCED_MAX_INSTANCE_LOAD))
    # TODO


import os
from jolly import env


class Config:
  ENV = ''

  def __init__(self):
    self.BALANCED_MAX_INSTANCE_LOAD = os.environ.get('BALANCED_MAX_INSTANCE_LOAD', 100)


class ProdConfig(Config):
  ENV = env.PROD


class StagingConfig(Config):
  ENV = env.STAGING


class DevConfig(Config):
  ENV = env.DEV


def get_config():
  config_class = globals().get('{}Config'.format(env.env().capitalize()))
  return config_class() if config_class else None


config = get_config()


import sys
from logging import INFO, FileHandler, StreamHandler, getLogger
from jolly.definitions import jolly
from jolly.env import is_prod

# Create Jolly logger.
logger = getLogger(jolly)
logger.addHandler(StreamHandler(sys.stdout) if is_prod() else FileHandler('{}.log'.format(jolly)))
logger.setLevel(INFO)


from .dynamo_db import dynamo_db, DynamoError
from .table import Table
from jolly.exp import JollyDBException
from jolly.logger import logger


class DynamoTable(Table):

  LAST_EVALUATED_KEY = 'LastEvaluatedKey'
  ITEMS_KEY = 'Items'

  def __init__(self, name):
    super(DynamoTable, self).__init__(name)

  def _new_table(self):
    return dynamo_db.Table(self.name)

  def get_all(self):
    try:
      resp = self._table.scan() or {}
      items = resp.get(self.ITEMS_KEY) or []

      while self.LAST_EVALUATED_KEY in resp:
        resp = self._table.scan(ExclusiveStartKey=resp[self.LAST_EVALUATED_KEY]) or {}
        items.extend(resp.get(self.ITEMS_KEY) or [])

      return items
    except DynamoError as e:
      logger.error('Error querying DynamoDB table {} for all items: {}'.format(self.name, e))
      raise JollyDBException(e)

  def create_item(self, **kwargs):
    try:
      return self._table.put_item(Item=dict(**kwargs))
    except DynamoError as e:
      logger.error('Error creating new DynamoDB item in table {}: {}'.format(self.name, e))
      raise JollyDBException(e)

  def update_item(self, **kwargs):
    try:
      return self._table.update_item(**kwargs)
    except DynamoError as e:
      logger.error('Error updating DynamoDB item in table {}: {}'.format(self.name, e))
      raise JollyDBException(e)


from .jolly_instances_table import JollyInstancesTable
from .jolly_resources_table import JollyResourcesTable

# Create table references.
jolly_instances_table = JollyInstancesTable()
jolly_resources_table = JollyResourcesTable()


def create_jolly_resource(jolly_resource):
  return jolly_resources_table.create_item(jolly_resource)


def get_jolly_instances():
  return jolly_instances_table.get_all()


def update_jolly_instance_load(instance, load):
  return jolly_instances_table.update_load(instance, load)


class Table(object):

  def __init__(self, name):
    self.name = name
    self._table = self._new_table()

  def _new_table(self):
    raise NotImplementedError

  def get_all(self):
    raise NotImplementedError


from jolly.exp import JollyException


class AbstractJollyService(object):

  def __init__(self, *args, exp=JollyException, **kwargs):
    self.exp = exp

  def perform(self):
    raise NotImplementedError

  def abort(self, *args, **kwargs):
    raise self.exp(*args, **kwargs)



import os

DEV = 'dev'
STAGING = 'staging'
PROD = 'prod'

ENV = os.environ.get('ENV') or DEV


def env():
  return ENV


def is_dev():
  return ENV == DEV


def is_staging():
  return ENV == STAGING


def is_prod():
  return ENV == PROD



from setuptools import setup, find_packages

setup(name='jolly',
      version='0.0.1',
      description='Jolly Python Client',
      url='https://github.com/whittlbc/jolly-py',
      author='Ben Whittle',
      author_email='benwhittle31@gmail.com',
      license='MIT',
      packages=find_packages(),
      install_requires=[
        'boto3==1.17.22',
      ],
      zip_safe=False)


class JollyInstance(object):

  def __init__(self, instance, ip, load=0):
    self.instance = instance
    self.ip = ip
    self.load = load


class JollyResource(object):

  MESSAGE_TYPE = 'message'
  CONVERSATION_TYPE = 'conversation'
  resource_types = [MESSAGE_TYPE, CONVERSATION_TYPE]

  @property
  def load(self):
    return {
      self.MESSAGE_TYPE: 1,
      self.CONVERSATION_TYPE: 5,
    }.get(self.resource_type)

  def __init__(self, resource_id, resource_type, instance):
    self.resource_id = resource_id
    self.resource_type = resource_type
    self.instance = instance



class JollyException(BaseException):

  def __init__(self, err='', **kwargs):
    super(JollyException, self).__init__(
      'Jolly exception: {} -- ({})'.format(err, kwargs))


class JollyDBException(JollyException):

  def __init__(self, err='', **kwargs):
    super(JollyDBException, self).__init__(
      'Jolly DB operation failed with exception: {} -- ({})'.format(err, kwargs))


class CreateJollyResourceException(JollyException):

  def __init__(self, err='', **kwargs):
    super(CreateJollyResourceException, self).__init__(
      'Creating Jolly resource failed with exception: {} -- ({})'.format(err, kwargs))


class SelectJollyInstanceException(JollyException):

  def __init__(self, err='', **kwargs):
    super(SelectJollyInstanceException, self).__init__(
      'Selecting Jolly instance failed with exception: {} -- ({})'.format(err, kwargs))


class UpdateJollyInstanceException(JollyException):

  def __init__(self, err='', **kwargs):
    super(UpdateJollyInstanceException, self).__init__(
      'Updating Jolly instance failed with exception: {} -- ({})'.format(err, kwargs))
